// video_pong
// play pong with your friend via a video camera
// by Joshua Minor - http://lux.vu/
// built with processing (www.processing.org)
//
// v1, 4/2003 basic pong game using difference-based motion detection
// v2, 5/2003 removed the need to manually sample a reference frame
// v3, 7/2006 upgraded to latest processing
// v4, 8/2006 we now compensate for CPU speed.
//            + adjusted collision response so vy of paddle is added to the ball
// v5, 4/2007 restyled to be closer to the original pong's visuals
//            + updated to work with Processing 0124

/*
    Copyright 2003,2006,2007 Joshua Minor
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// once upon a time

import processing.video.*; 

// flop the video so it appears like a mirror
boolean flop = true;

// frame counter for animation
int frame = 0;
float now, then;

// this trigger causes a frame to be saved to disk
boolean takeSnap = false;

PFont font;
int bigFontSize = 60;
int littleFontSize = 20;

// video capture & reference image for motion detection
Capture capture;
PImage current, reference;
boolean haveVideo = false;
String errorMessage;

// you can adjust this tollerance if the video is noisy
int tollerance = 40;
int showtollerance = 0;

// a = the left player
// b = the right player
Thing ball, paddlea, paddleb, detectora, detectorb;
int pointsa, pointsb;
int countdown;

// how many points to play until
int playto = 5;

void setup()
{
  size(640,480,P3D);

  try {
    println("Available video devices:");
    String devices[] = Capture.list();
    println(devices);
    // we just pick the default one
    capture = new Capture(this, width, height, 15);
    haveVideo = true;
  }
  catch( java.lang.Throwable ex ) {
    // try to catch any sort of error from Capture and inform the user
    errorMessage = ex.toString();
    println("Error accessing video camera.");
    println(errorMessage);
  }
  background(0);
  frameRate(60);

  font = loadFont("Futura-Medium.vlw");
  textFont(font, bigFontSize);

  resetAll();
}

void resetAll()
{
  // scale the objects to match the resolution of the screen
  float thickness = width/30.;

  // new Thing(x, y, vx, vy, width, height)
  ball    = new Thing(width/2-thickness/2, height/2-thickness/2, 0, 0, thickness, thickness);
  paddlea = new Thing(thickness,           0, 0, 0, thickness, thickness*5);
  paddleb = new Thing(width-thickness*2,   0, 0, 0, thickness, thickness*5);
  detectora = new Thing(0,0,0,0,0,0);
  detectorb = new Thing(0,0,0,0,0,0);

  pointsa = 0;
  pointsb = 0;

  reference = null;

  resetBall();

  now = millis();
}

void resetBall() {
  ball.c = color(255,255,255);

  // move to the center
  ball.moveTo( width/2 - ball.width/2, height/2 - ball.height/2 );

  // pick a random direction (left or right)
  int direction = 1;
  if( random(1) < 0.5 ) direction = -1;

  // pick a random velocity to start with
  ball.setVelocity( direction*(random(width/3)+width/6), random(width/6)-width/6 );

  // begin counting down to the start of the game
  countdown = 3*30;
}

void draw()
{
  // count frames
  frame++;

  // note how much time has passed since our last loop
  then = now;

  // wait for at least 1/100th of a second to pass
  float elapsed = 0;
  while( elapsed < 1.0/100.0 ) {
    now = millis();

    // compute elapsed time in seconds
    elapsed = (now - then) / 1000.0f;
  }

  // slow the game down if the computer can't keep up a reasonable frame rate
  // by limiting the elapsed time to 1/12th of a second (12fps)
  if( elapsed > 1.0/12.0 ) elapsed = 1.0/12.0;

  if( haveVideo ) {
    if( capture.available() ) {
      // keep a copy of the old one for reference
      reference = current;
      // get the new frame
      capture.read();
      current = capture.get();
    }

    if( flop ) {
      pushMatrix();
      scale(-1,1);
      image(current,-width,0,width,height);
      popMatrix();
    }
    else{
      image(current,0,0,width,height);
    }
  }
  else{
    background(0);
    textFont(font, littleFontSize);
    centerText( "Error accessing video camera", height/2 - 15, color(255) );
    centerText( errorMessage, height/2 + 15, color(255) );
    textFont(font, bigFontSize);
    return;
  }

  // if we just adjusted the tollerance, show the amount for a few seconds
  if( showtollerance > 0 ) {
    textFont(font, littleFontSize);
    centerText( "Tollerance:"+tollerance, height/2, color(255) );
    textFont(font, bigFontSize);
    showtollerance--;  
  }

  // do we have both current and reference video available?
  if( reference != null ) {
    // look for movement between the reference frame and the current frame
    // in the left and right zones, one for each player
    boolean foundA = detectora.findMovement( width - (int)(paddlea.x + paddlea.width), width - (int)(paddlea.x), 0, height-1 );
    boolean foundB = detectorb.findMovement( width - (int)paddleb.x, width - (int)(paddleb.x + paddleb.width), 0, height-1 );

    // adjust the targets by the size of the paddles
    float ay,by;
    ay = detectora.y - paddlea.height/2;
    by = detectorb.y - paddleb.height/2;

    // instead of jumping directly to the detected point, lets move towards it gradually
    // this keeps the motion smooth.
    float gain = 0.5;    // adjust this to tradeoff between speed & smoothness
    float olday = paddlea.y;
    float oldby = paddleb.y;
    paddlea.y = paddlea.y*(1.0-gain) + ay*gain;
    paddleb.y = paddleb.y*(1.0-gain) + by*gain;
    paddlea.vy = (paddlea.y - olday) / elapsed;
    paddleb.vy = (paddleb.y - oldby) / elapsed;
    paddlea.c = color(255,255,255);
    paddleb.c = color(255,255,255);
    noStroke();

    // draw the center line
    int x, y;
    x = width/2 - 2;
    for( y=0; y<height; y+=ball.height ) {
      fill(0,0,0,127);
      rect(x+5, y+5, 4, ball.height/2);
      fill(255,255,255,127);
      rect(x, y, 4, ball.height/2);
    }

    // did someone win?
    if( pointsa >= playto || pointsb >= playto ) {
      centerText( "Winner!", height/2, color(255) );

      if( countdown == 0 ) {
        countdown = 3*30;
      }
      else if( countdown == 1 ) {
        resetAll();
      }
      countdown--;

      if( (countdown & 2) == 0 ) {
        if( pointsa > pointsb ) {
          paddlea.c = color(255,255,0);
        }
        else{
          paddleb.c = color(255,255,0);
        }
      }
    }
    else{
      if( countdown > 0 ) {
        String message;
        if( countdown > 2*30 ) {
          message = "Ready";
        }
        else if( countdown > 1*30 ) {
          message = "Set";
        }
        else{
          message = "Go!";
        }

        if( message != null ) {
          centerText( message, (int)(height/2 - ball.height*2 - 30), color(255) );
        }
        countdown--;
      }

      if( countdown < 1*30 ) {
        ball.move(elapsed);

        // bounce off the paddles
        if( ball.vx < 0 && ball.intersects(paddlea) ) {
          ball.vx = ball.vx * -1.05;
          // most pong games bounce the ball off the paddle as though it was rounded
          // like this:  ball.vy = ((ball.y - paddlea.y) / paddlea.height) * width - width/2;
          // but we do it differently by taking a fraction of the paddle's velocity and adding it
          // to the ball's.  this makes it feel more like hitting the ball with the paddle
          ball.vy += paddlea.vy / 2;
        }
        else
          if( ball.vx > 0 && ball.intersects(paddleb) ) {
            ball.vx = ball.vx * -1.05;
            //ball.vy = ((ball.y - paddleb.y) / paddleb.height) * width - width/2;
            ball.vy += paddleb.vy / 2;
          }

        // bounce off the top & bottom
        if( ball.y < 0 || ball.y+ball.height > height ) {
          ball.vy = -ball.vy;
          ball.y += ball.vy * elapsed;
        }

        // did someone score?
        if( ball.x+ball.width < 0 ) {
          pointsb++;
          resetBall();
        }
        else if( ball.x > width ) {
          pointsa++;
          resetBall();
        }
      }
    }

    // draw the left paddle
    paddlea.draw();

    // draw the ball in the middle so the drop-shadows overlap nicely
    // don't draw the ball if someone just won
    if( pointsa < playto && pointsb < playto ) {
      ball.draw();
    }

    // draw the right paddle
    paddleb.draw();

    // draw the scores atop everything
    myText( ""+pointsa, (int)paddlea.width*3, bigFontSize+10, paddlea.c );
    myText( ""+pointsb, (int)(width-paddleb.width*3-textWidth(""+pointsb)), bigFontSize+10, paddleb.c );
  }

  // take a snapshot?
  if( takeSnap ) {
    saveFrame();
    //takeSnap = false;
  }
}

void keyPressed()
{
  if( key == 's' || key == 'S' ) {
    takeSnap = true;
  }
  else if( key == '=' || key == '+' ) {
    tollerance++;
    if( tollerance > 255 ) tollerance = 255;
    showtollerance = 10;
  }
  else if( key == '-' || key == '_' ) {
    tollerance--;
    if( tollerance < 1 ) tollerance = 1;
    showtollerance = 10;
  }
  else{
    ;
  }
}

void mouseClicked()
{
  resetBall();
}

void centerText( String string, int y, color c )
{
  myText( string, (int)(width/2-textWidth(string)/2), y, c );
}

void myText( String string, int x, int y, color c )
{
  // draw text with a semi-transparent drop-shadow
  fill( 0,0,0,127 );
  text( string, x+5, y+5 );
  fill( c );
  text( string, x, y );
}

// paddle & ball data
class Thing {
  float x,y,vx,vy;
  float width,height;
  color c;
  public Thing(float x, float y, float vx, float vy, float w, float h) {
    this.x = x;
    this.y = y;
    this.vx = vx;
    this.vy = vy;
    this.width = w;
    this.height = h;
    this.c = color(255,255,255);
  }
  public void draw() {
    // draw with a semi-transparent drop-shadow
    fill(0,0,0,127);
    rect(x+5, y+6, width, height);
    fill(c);
    rect(x, y, width, height);
  }
  public void move(float elapsed) {
    x += vx * elapsed;
    y += vy * elapsed;
  }
  public void moveTo(float x, float y) {
    this.x = x;
    this.y = y;
  }
  public void setVelocity(float vx, float vy) {
    this.vx = vx;
    this.vy = vy;
  }
  public boolean intersects(Thing t) {
    if( x < t.x+t.width && x+width > t.x && y < t.y+t.height && y+height > t.y ) {
      return true;
    }
    return false;
  }
  // detect movement within a given bounding box
  boolean findMovement( int x1, int x2, int y1, int y2 )
  {
    int maxx = max(x1,x2);
    int minx = min(x1,x2);
    int maxy = max(y1,y2);
    int miny = min(y1,y2);
    int x,y, count=0;
    float accumy = 0;
    float accumx = 0;

    stroke(255,255,0);

    // scan the image looking for pixels that differ from the reference
    for( x=minx; x<=maxx; x+=2 ) {
      for( y=miny; y<=maxy; y+=2 ) {
        int index = y*current.width + x;
        color c1 = current.pixels[index];
        color c2 = reference.pixels[index];
        float r = abs(red(c1)-red(c2));
        float g = abs(green(c1)-green(c2));
        float b = abs(blue(c1)-blue(c2));
        if( r+g+b > tollerance ) {
          // this is sufficently different to qualify
          // mark it on screen - this makes the area sparkle
          point(current.width-x,y);
          // accumulate this pixel into a running average
          accumy += y;
          accumx += x;
          count++;
        }
      }
    }

    // did we find enough movement to be worth while?
    if( count > 50 ) {
      // move to the center of the movement
      moveTo(accumx/count, accumy/count);
      return true;
    }
    else{
      // we didn't find anything interesting
      return false;
    }
  }
}

// the end
